#include "httpd.h"
#include "http_config.h"
#include "http_main.h"
#include "http_protocol.h"
#include "http_log.h"
#include "ap_config.h"

/* PostgreSQL access functions */
#include "libpq-fe.h"

#include "co_utils.h"
#include "co_log.h"
#include "co_lib.h"
#include "co_defines_calls.h"
#include "co_defines.h"

#include "co_mod.h" /* include before co_js.h */

#include "co_submodule.h"

static const char *query_modules =
"
select C.class_name from co_sys_meta_method M, co_sys_meta_class C
where M.id_meta_class = C.id
 ;
";

/* comments on library/module handling see: mod_so.c */
static void unload_library(coffice_libinfo *libinfo)
{
    if (libinfo->dlh == NULL)
        return;

    /*dlclose (libinfo->dlh);*/
    ap_os_dso_unload(libinfo->dlh);

    libinfo->dlh = NULL;
    libinfo->name = NULL;
    libinfo->new_param = NULL;
    libinfo->set_param = NULL;
    libinfo->get_param = NULL;
}

void *coloadsubmodule(
		server_rec *s,
		const char *name /* name of the submodule */)
{
	coffice_libinfo *libinfo;
	void *dlh;
	pool *p, *local_pool;
	coffice_server_conf *coffice_sconf;

	p = cou_getCOModulePool();
	local_pool = ap_make_sub_pool(p);
	coffice_sconf = cox2_getCOfficeConfig( s );
	
	/* you can do: dlh = ap_os_dso_open(...) - it is more portable
	 but there is no need of exporting symbols (ap_os_dso_open uses RTLD_NOW | RTLD_GLOBAL)*/
	dlh = dlopen(
		(char*)ap_pstrcat( local_pool, cou_getCOModulePath(), "/co_", name, ".so", NULL),
		RTLD_NOW
	    );
	if (dlh) {
		libinfo = ap_push_array( cou_getCOLoadedLibs() );
		libinfo->name = ap_pstrdup( p, name);
		libinfo->dlh = dlh;
		libinfo->new_param = ap_os_dso_sym( dlh, 
		    ap_pstrcat( local_pool, "coj_", name, "_new_param", NULL )
		);
		libinfo->set_param = ap_os_dso_sym( dlh, 
		    ap_pstrcat( local_pool, "coj_", name, "_set_param", NULL )
		);
		libinfo->get_param = ap_os_dso_sym( dlh, 
		    ap_pstrcat( local_pool, "coj_", name, "_get_param", NULL )
		);
		ap_register_cleanup( p, libinfo,
                     (void (*)(void*))unload_library, ap_null_cleanup);
	    ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, 
			"coloadsubmodule: loaded module %s",	name
		);
		ap_destroy_pool(local_pool);
		return dlh;
	}
	else {
	    ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, s, 
		"coloadsubmodule: cannot open module %s\nLOADING ERROR: dlopen failed\nREASON: %s",
		name,dlerror());
	}
	ap_destroy_pool(local_pool);
	return NULL;
}

void *cou_loadSubmodule(
        co_context *cnx,
        const char *name /* name of the submodule */)
{
	return coloadsubmodule( cnx->request->server, name );
}