#include "types.t"
#include "defines.t"

class generic_association
{
   area "sys";

   container_class '';
   container_column_name '';   

#include "check_acl.t"
    
   inherits generic_acl;

   uses "generic";

#include "generic_acl_props.t" 

include_c "../coffice/mod_coffice.h";

	method showFormAssociation
	{
		doc "this method shows form for creating/browsing associations of two classes";
		doc "Returns: nothing";

		id_permission __PERMISSION_SELECT;
		sec_type __METHOD_SEC_TYPE_CLASS;

		__PARAMS_HEADER

		param column_first_class
		{
			type CO_STRING;
			optional 1;
			default "column_first_class=NULL";
			doc "this is the name of the column containing reference to the first class";
		};
 
		param report_first_class
		{
			type CO_STRING;
			optional 1;
			default "report_first_class=NULL";
			doc "this is the name of the report for selecting records from the first class";
		};
 
		param column_second_class
		{
			type CO_STRING;
			optional 1;
			default "column_second_class=NULL";
			doc "this is the name of the column containing reference to the second class";
		};
 
		param report_second_class
		{
			type CO_STRING;
			optional 1;
			default "report_second_class=NULL";
			doc "this is the name of the report for selecting records from the second class";
		};
 
		declare language_c
		{
			C "const char* filename;";	
			C "const char* root_path;";	
			C "const char* prop_foreign1;";	
			C "const char* prop_foreign2;";	
		};  
 
		body language_c
		{
			C 
			%%

            __SQL_GET_STRING( prop_foreign1, "SELECT prop_foreign FROM co_%s_meta_prop WHERE
 class_name='%s' and prop_name='%s'", par_area, par_class, column_first_class)

            __SQL_GET_STRING( prop_foreign2, "SELECT prop_foreign FROM co_%s_meta_prop WHERE
 class_name='%s' and prop_name='%s'", par_area, par_class, column_second_class)

			//cou_putsHTML("<FRAMESET rows=\"50,50%,*\">");			
			root_path = cou_getCORootPath( cnx );
			filename = ap_psprintf(p, "%s/generic_association/showFormAssociation.asp", root_path);
			__PUT_ASP_FROM_FILE(filename)

			%%
			;
		};
	}; // end of showFormAssociation

	method showFormMenu
	{
		doc "this method creates object and sets ACLs taken from acl_class, also sets props as given in create_params";

		id_permission __PERMISSION_SELECT;
		sec_type __METHOD_SEC_TYPE_CLASS;

		__PARAMS_HEADER

		param menu_name
		{
			type CO_STRING;
			optional 1;
			default "menu_name=\"default\"";
			doc "this is the name of the menu ('default' by default)";
		};
 
		param menu_type
		{
			type CO_STRING;
			optional 1;
			default "menu_type=\"multi\"";
			doc "this is the type (single, multi) of the menu ('multi' by default)";
		};

		declare language_c
		{
			C "const char* filename;";	
			C "const char* root_path;";	
		};  
 
		body language_c
		{
			C 
			%%

			root_path = cou_getCORootPath( cnx );
			filename = ap_psprintf(p, "%s/menu/showMenu.asp", root_path);
			__PUT_ASP_FROM_FILE(filename)
	
			%%
			;
		};
	}; // end of showFormAssociation


	method insertAssociations
	{
		doc "this method creates object and sets ACLs taken from acl_class, also sets props as given in create_params";
		doc "Returns: ???";

		id_permission __PERMISSION_CREATE;
		sec_type __METHOD_SEC_TYPE_CLASS;

		__PARAMS_HEADER_WITH_CALLBACK

		param ids_first_class
		{
			type CO_ARRAY_OF_STRINGS;
			optional 1;
			default "ids_first_class=NULL";
			doc "this is array of ids belonging to the first class";
		};
 
		param prop_first_class
		{
			type CO_STRING;
			optional 1;
			default "prop_first_class=NULL";
			doc "this is name of the property storing ids belonging to the first class";
		};
 
		param ids_second_class
		{
			type CO_ARRAY_OF_STRINGS;
			optional 1;
			default "ids_second_class=NULL";
			doc "this is array of ids belonging to the second class";
		};
 
		param prop_second_class
		{
			type CO_STRING;
			optional 1;
			default "prop_second_class=NULL";
			doc "this is name of the property storing ids belonging to the second class";
		};
 
		declare language_c
		{
			C "int NIds1=0, NIds2=0;";	
			C "int i,j;";
			C "int id_new;";
			C "const char *str_id1, *str_id2;";	
			C "co_array_of_pointers * create_params;";
			C 'co_recordset * inserted_corec;';
		};  
 
		body language_c
		{
			C 
			%%
			inserted_corec = cor_CreateEmpty(cnx);
			NIds1 = coa_GetElementCount(cnx, ids_first_class);
			NIds2 = coa_GetElementCount(cnx, ids_second_class);
			for(i=0; i<NIds1; i++)
			{
				str_id1 = coa_GetStringElement(cnx, ids_first_class, i);
				for(j=0; j<NIds2; j++)
				{
					str_id2 = coa_GetStringElement(cnx, ids_second_class, j);
					// create new record, then select it and add to result
					create_params = coa_Create( cnx );
					coa_AppendElement( cnx, create_params, prop_first_class);
					coa_AppendElement( cnx, create_params, str_id1);
					coa_AppendElement( cnx, create_params, prop_second_class);
					coa_AppendElement( cnx, create_params, str_id2);
					@co_lcall(par_area, par_class, "createAndSet", \
						par_area=>par_area, par_class=>par_class, par_id=>par_id, \
						create_params=>create_params, callback=>'""');
					__GET_RETURNED_INTEGER(id_new);
					__CREATE_QUERY("SELECT * FROM co_%s_%s WHERE id=%d", 
						par_area, par_class, id_new)
					__SQL_GET_RESULT
					cor_AppendRecordset(cnx, inserted_corec, corec);
					__CLEAR_RESULT
				}
			}
			__CREATE_JS_FROM(inserted_corec)
			__RETURN_RECORDSET(inserted_corec)
			%%
			;
		};
	}; // end of insertAssociations


};  // end of class generic_association
