#include "types.t"
#include "defines.t"

// CLASS GENERIC_VIEW

class generic_view
{
  area "sys";

  inherits 0;

#include "check_acl_view.t"

  container_class '';
  container_column_name '';

  uses "generic";

  include_c "http_protocol.h";
  include_c "http_main.h";
  include_c "co_session.h";

  default_form "";

 // PROPERTIES

// METHODS

//TODO: What if not found?

  method getByProp
  {
   id_permission __PERMISSION_SELECT;
   sec_type __METHOD_SEC_TYPE_CLASS;
 
   doc "This method retrieves data for object which has prop_name == prop_val. When many objects satysfy this conditions all are returned"
   {
       pl "Ta metoda zwraca dane dla obiektu, ktora ma prop_name == prop_val. W przypadku, gdy wiele obiektow spelnia ten warunek zwraca wszystkie te obiekty.";
   };
   
    __PARAMS_HEADER_WITH_CALLBACK
    param prop_name 
    {
      type CO_STRING;
      doc "Name of prop used to find the object"
      {
         pl "Nazwa prop uzywanego do znalezienia obiektu";
      };

    };
    param prop_val 
    { 
      type CO_STRING;
      doc "Value used to find the object"
      {
         pl "Wartosc uzywana do znalezienia obiektu";
      };
    };
    
    body language_c
      {
         C 
         %%
          __CREATE_QUERY("SELECT * FROM cov_%s_%s WHERE %s='%s'" , par_area, par_class, prop_name, prop_val)
         __GET_RESULT_AND_CREATE_JS
         %%
         ; 
      };  
								    
  };

//TODO: What if not found?

  method selectByProp
  {
   id_permission __PERMISSION_SELECT;
   sec_type __METHOD_SEC_TYPE_CLASS;

   doc "This method retrieves data for objects which has prop_name LIKE prop_val"
   {
       pl "Ta metoda zwraca dane obiektow spelniajacych warunek prop_name LIKE prop_val";
   };
   
    __PARAMS_HEADER_WITH_CALLBACK
    param prop_name 
    {
      type CO_STRING;
      doc "Name of prop used to select objects"
      {
         pl "Nazwa prop uzywanego do wybrania obiektow";
      };

    };
    param prop_val 
    { 
      type CO_STRING;
      doc "Value used to select the object"
      {
         pl "Wartosc uzywana do wybrania obiektow";
      };
    };
 
    body language_c
      {
         C 
         %%
          __CREATE_QUERY("SELECT * FROM cov_%s_%s WHERE %s LIKE '%s'" , par_area, par_class, prop_name, prop_val)
         __GET_RESULT_AND_CREATE_JS
         %%
         ; 
      };  
								    
  };


  method selectX
  {
    doc "Gets recordset of objects on given filter rule";
    sec_type __METHOD_SEC_TYPE_CLASS;
    
    id_permission __PERMISSION_SELECT;

    __PARAMS_HEADER_WITH_CALLBACK    

    param get_class_names 
    {
      doc "In this param we pass names of classes from wich we are reading data"
      {
         pl "W tym parametrze przekazujemy liste nazw klas, z ktorych bedziemy czytac dane";
      };
      type CO_ARRAY_OF_STRINGS;
    };

    param get_prop_names
    {
      doc 
       %%
         In this param we pass names of fields wich we want to get. You can use "outer join" notation: field_name>related_field_name>class_name>field_name
         For example: id_role>id>role>name -- gets the name of role linked by relation class_name.id_role = role.id
         This works like outerjoin, so if there is no match you get empty value.
         Number of fields must be equal to number of class_names
       %%
      {
         pl 
         %%
          W tym parametrze przekazujemy liste nazw pol, ktore chcemy odczytac. 
          Mozna uzywac notacji "outer join": nazwa_pola>nazwa_pola_zwiazanego>nazwa_klasy>nazwa_pola
          Na przyklad: id_role>id>role>name -- odczytuje nazwe roli powiazana relacja class_name.id_role = role.id
          Odczytywanie dziala jak outer join, czyli w przypadku braku pasujacego rekordu jest zwracana pusta wartosc
          Liczba pol musi byc taka sama jak liczba nazw klas
         %%
         ;
      };
      type CO_ARRAY_OF_STRINGS;
    };
 
    param get_prop_aggregates
    {
      doc 
       %%
         In this param we pass names of aggregation function to be used to get results (like SUM, MIN, MAX, COUT, AVG or other defined)
         In selects with aggregates you can not use order, limit nor offset (are ignored) 
       %%
      {
         pl 
         %%
          W tym parametrze przekazujemy nazw funkcji agregujcej, ktra bdzie uzyta do wyliczenia wyniku (SUM, MIN, MAX, COUT, AVG lub inne zdefiniowane)
          W selektach z agregacj nie mona uywac order, limit i offset (s ignorowane)
         %%
         ;  
      };
      type CO_ARRAY_OF_STRINGS;
      optional 1;
      default "get_prop_aggregates = NULL;";
    };

    param filter_class_names
    {
      type CO_ARRAY_OF_STRINGS;
      doc "This is list of class names that we use to filter data"
      {
         pl "To jest lista nazw klas uzywanych do filtrowania danych";
      };
    };

    param filter_prop_names 
    {
      type CO_ARRAY_OF_STRINGS;
      doc "This is list of field names that we use to filter data. Length of this list must be equal to length of filter_class_names"
      {
         pl "To jest lista nazw pol uzywanych do filtrowania danych. Dlugosc tej listy musi byc taka sama jak dlugos listy filter_class_names";
      };
    };

    param filter_operators 
    {
      type CO_ARRAY_OF_STRINGS;
      doc "This is list of operators (LIKE, =, <, >, etc) that we use to filter data. Length of this list must be equal to length of filter_class_names.IMPORTANT: LIKE is case-insensitive here"
      {
         pl "To jest lista operatorow pol (LIKE, =, <, >, itd.) uzywanych do filtrowania danych. ";
         pl "Dlugosc tej listy musi byc taka sama jak dlugos listy filter_class_names";
         pl "LIKE nie rozrnia maych i duych liter (inaczej ni w psql-u)";
      };
    };

    param filter_values 
    {
      type CO_ARRAY_OF_STRINGS;
      doc "This is list of values that we use to filter data. Length of this list must be equal to length of filter_class_names"
      {
         pl "To jest lista wartosci uzywanych do filtrowania danych. Dlugosc tej listy musi byc taka sama jak dlugos listy filter_class_names";
      };
    };
 
    param filter_r_class_names 
    {
      doc "This param is used to compare value of one field with another field (not constant). In this param you pass class_names of fields to be used in comparasion"
      {
         pl "Ten parametr jest wykorzystywany przy porownywaniu wartosci jednego pola z innym polem. W tym parametrze podane sa nazwy klas pol do porownywania";
      };
      optional 1;
      default "filter_r_class_names = NULL;";
      type CO_ARRAY_OF_STRINGS;
    };

    param filter_r_prop_names 
    {
      doc "This param is used to compare value of one field with another field (not constant). In this param you pass prop_namea of fields to be used in comparasion"
      {
         pl "Ten parametr jest wykorzystywany przy porownywaniu wartosci jednego pola z innym polem. W tym parametrze podane sa nazwy pol do porownywania";
      };

      optional 1;
      default "filter_r_prop_names = NULL;";
      type CO_ARRAY_OF_STRINGS;
    };

    param filter_compare 
    {
      doc 'When this param is == "1" then we compare values of one field with another field (not constant)'
      {
         pl 'Kiedy ten parametr jest rowny "1" to jest wykonywane porownanie wartosci pol z innymi polami (a nie ze stalymi)';
      };
      optional 1;
      default "filter_compare = NULL;";
      type CO_ARRAY_OF_STRINGS;
    };
 
    param order_by_class_names
    {
      doc 'This are names of classes from which we take sort columns'
      {
         pl 'W tym parametrze podajemy nazwy klas dla kolumn, wedlug ktorych beda posortowane dane';
      };
      optional 1;
      default "order_by_class_names = NULL;";
      type CO_ARRAY_OF_STRINGS;
    };
 
    param order_by_prop_names
    {
      doc 'This are names of sort columns'
      {
         pl 'W tym parametrze podajemy nazwy kolumn, wedlug ktorych beda posortowane dane';
      };
      optional 1;
      default "order_by_prop_names = NULL;";
      type CO_ARRAY_OF_STRINGS;
    };
 
    param order_by_desc
    {
      doc 'If this param == "1", then we sort values on this position descending. You can provide less values then order_by_class_names, default order is ascending'
      {
         pl 'Na pozycjach, gdzie parametr ten jest rowny "1" sortujemy w porzadku malejacym. Mozna podac mniej parametrow niz order_by_class_names, domyslnie sortowanie jest w kolejnosci rosnacej';
      };
      optional 1;
      default "order_by_desc = NULL;";
      type CO_ARRAY_OF_STRINGS;
    };

   param limit
   {
     doc 'In this param we set limit number of records that we want in return. If == -1 (default) then there is no limit'
     {
        pl 'Ten parametr okresla limit ilosci rekordow, jakie zostana zwrocone. Jesli jest rowny -1 (domyslnie) to nie ma ograniczen.';
     };
     optional 1;
     default "limit = -1";
     type CO_INTEGER;
   };
 
   param offset
   {
     doc 'In this param we set offset number of records that we want in return.'
     {
        pl 'Ten parametr okresla ile rekordow na poczatku mamy pominac (offset).';
     };
     optional 1;
     default "offset = 0";
     type CO_INTEGER;
   };
 
    param group_by_class_names
    {
      doc 'This is the list of class names to group-by return values'
      {
         pl 'To jest lista nazw klas, wzgldem ktrych bd grupowane wyniki';
      };
      optional 1;
      default "group_by_class_names = NULL;";
      type CO_ARRAY_OF_STRINGS;
    };


    param group_by_prop_names
    {
      doc 'This is the list of props to group-by return values'
      {
         pl 'To jest lista prop, wzgldem ktrych bd grupowane wyniki';
      };
      optional 1;
      default "group_by_prop_names = NULL;";
      type CO_ARRAY_OF_STRINGS;
    };

   
    declare language_c
    {
      C
      %%
      int get_col_num, filter_col_num;
      int sql_from_start,i;
      char column_value_coded[FIELDBUFSIZE];
      char column_value_js_encoded[FIELDBUFSIZE];
      char *as_field_name;
      char start_string[10];
      char *fields[MAXFIELDNUMBER];
      char * in_fields;
      int field_elems = 0;
      int j;
      int more_compare;
      int more_desc;
      int NGetNames;
      char * cur_class_name; 

      int NFilterNames;
      char * cur_filter_name; 

      int NElems;
      char * cur_elem;

      %%
      ;
    };
      
    body language_c
    {
      C
      %%

      memset(column_value_coded,0,FIELDBUFSIZE);
      memset(column_value_js_encoded,0,FIELDBUFSIZE);

      __SEND_MESG_DEBUG("ENTERING SelectX")

      __CREATE_QUERY("SELECT ");

     NGetNames = coa_GetElementCount(cnx, get_class_names);

     for(i=0; i<NGetNames; i++)
     {
         char * as_fields;
         if(i==0)
         {
           snprintf(start_string,10,"\n");
         }
         if(i==1)
         {
           snprintf(start_string,10," , \n");
         } 

         __SEND_MESG_DEBUG("We are going to parse field: %s", coa_GetStringElement(cnx, get_prop_names, i))

         in_fields = ap_pstrdup(p, coa_GetElement(cnx, get_prop_names, i));
         as_fields = ap_pstrdup(p, coa_GetElement(cnx, get_prop_names, i));

         field_elems = cou_split(in_fields, fields, ">", MAXFIELDNUMBER);

         __ASSERT(field_elems>0 , "In selectX error in parsing field: %s", in_fields)

         cur_class_name = coa_GetElement(cnx, get_class_names, i);
         as_field_name = ap_psprintf(p, "cov_%s_%s.\"%s\"", par_area, cur_class_name, fields[0]);

         if(field_elems>1)
         {
           for(j=1; j<field_elems; j+=3)
           {
              as_field_name = ap_psprintf(p, "%s(lhlo_%s_%s_%s(%s))", fields[j+2], par_area, fields[j+1], fields[j], as_field_name);
           }
           if(get_prop_aggregates!=NULL)
           {
              as_field_name = ap_psprintf(p, "%s(%s)", (char*)coa_GetElement(cnx, get_prop_aggregates, i), as_field_name);
           }
           __APPEND_TO_QUERY(" %s %s AS \"%s\"", start_string, as_field_name, as_fields);
         }
         else
         { 
           if(get_prop_aggregates!=NULL)
           {
              as_field_name = ap_psprintf(p, "%s(%s)", (char*) coa_GetElement(cnx, get_prop_aggregates, i), as_field_name);
           }
           if(strcmp(fields[0],"*")==0)
           { 
                __APPEND_TO_QUERY(" %s %s ", start_string, as_field_name);
           }else
           {
                __APPEND_TO_QUERY(" %s %s AS \"%s\"", start_string, as_field_name, as_fields);
            }
         } //ifelse 
      } // for get_class_names
      
      __SEND_MESG_DEBUG("We have field part of SQL: %s", SQLString)

      sql_from_start = strlen(SQLString);
      __APPEND_TO_QUERY(" FROM cov_%s_%s ", par_area, (char *) coa_GetElement(cnx, get_class_names, 0));
      for(i=1; i<NGetNames;i++)
      { 
       cur_class_name = coa_GetElement(cnx, get_class_names, i);
        /* all table names we are taking only once*/
       if(0==strstr(SQLString+sql_from_start, cur_class_name))
       { 
         __APPEND_TO_QUERY(" , cov_%s_%s ", par_area, cur_class_name);
       }
      }

      __SEND_MESG_DEBUG("We are after get_class_names")

      NFilterNames = coa_GetElementCount(cnx, filter_class_names);
      for(i=0; i<NFilterNames; i++)
      { 
        cur_filter_name = coa_GetElement(cnx, filter_class_names, i); 
        /* all table names we are taking only once*/
       if(0==strstr(SQLString+sql_from_start, cur_filter_name))
       { 
         __APPEND_TO_QUERY(" , cov_%s_%s ", par_area, cur_filter_name);
       }
      }

      __SEND_MESG_DEBUG("We are after filter_class_names")

      if(filter_r_class_names!=NULL)
      {
        NElems = coa_GetElementCount(cnx, filter_r_class_names); 
        for(i=0; i<NElems; i++)        
        {
          cur_elem = coa_GetElement(cnx, filter_r_class_names, i);  
          /* all table names we are taking only once*/
          if(0==strstr(SQLString+sql_from_start, cur_elem) && strlen(cur_elem)>0)
          { 
             __APPEND_TO_QUERY(" , cov_%s_%s ", par_area, cur_elem);
          }
        }
      }
      
      __SEND_MESG_DEBUG("We are after filter_r_class_names")

      more_compare = 0;
      if(filter_compare != NULL) more_compare = 1;
      NElems = coa_GetElementCount(cnx, filter_prop_names);
      for(i=0; i<NElems; i++)
      {
         char * curr_filter_class_names  = coa_GetElement(cnx, filter_class_names, i);
         char * curr_filter_prop_names   = coa_GetElement(cnx, filter_prop_names, i);
         char * curr_filter_operators    = coa_GetElement(cnx, filter_operators, i);
         char * curr_filter_values       = coa_GetElement(cnx, filter_values, i);

         // these can be NULL
         char * curr_filter_compare      = (filter_compare     == NULL ? NULL : coa_GetElement(cnx, filter_compare, i));
         char * curr_filter_r_class_names= (filter_r_class_names == NULL ? NULL : coa_GetElement(cnx, filter_r_class_names, i));
         char * curr_filter_r_prop_names = (filter_r_prop_names  == NULL ? NULL : coa_GetElement(cnx, filter_r_prop_names, i));
   
         if(i==0)
         {
           snprintf(start_string,10," WHERE \n");
         }
         if(i==1)
         {
           snprintf(start_string,10, " AND \n");
         }
         if(more_compare && curr_filter_compare==NULL) more_compare=0;
         if(more_compare && strcmp(curr_filter_compare, "1")==0)
         {
           if(strcmp(curr_filter_operators,"LOOSE")==0)
           {
               __APPEND_TO_QUERY(" %s (cov_%s_%s.%s = cov_%s_%s.%s OR cov_%s_%s.%s = '' ) \n", start_string, par_area,
                  curr_filter_class_names, curr_filter_prop_names, par_area, curr_filter_r_class_names, curr_filter_r_prop_names,
                  par_area, curr_filter_r_class_names, curr_filter_r_prop_names);
           }
           else
           {
	       __APPEND_TO_QUERY(" %s cov_%s_%s.%s %s cov_%s_%s.%s ", start_string, par_area,
                 curr_filter_class_names, curr_filter_prop_names, curr_filter_operators, par_area, curr_filter_r_class_names, curr_filter_r_prop_names);
	   }
         }
         else
         {
				
				field_elems = cou_split(curr_filter_prop_names, fields, ">", MAXFIELDNUMBER);

				__ASSERT(field_elems>0 , "In selectX error in parsing filter props: %s", curr_filter_prop_names)

		        as_field_name = ap_psprintf(p, "cov_%s_%s.%s", par_area, curr_filter_class_names, fields[0]);

				if(field_elems>1)
				{
					for(j=1; j<field_elems; j+=3)
					{
						as_field_name = ap_psprintf(p, "%s(lhlo_%s_%s_%s(%s))", fields[j+2], par_area, fields[j+1], fields[j], as_field_name);
					}
				}
            if (strcasecmp(curr_filter_operators,"IN")==0)
		    {
				// TODO: check that curr_filter_values is list of values. 
		      __APPEND_TO_QUERY(" %s %s IN %s ", start_string, as_field_name, curr_filter_values);
            }
		    else if (strcasecmp(curr_filter_operators,"LIKE")==0) // case-insensitive LIKE
			{
		      // we have to optimize this (evaluate upper('%s') in c code)
		      __APPEND_TO_QUERY(" %s upper(%s) LIKE upper(text(\'%s\')) ", start_string, as_field_name, curr_filter_values);
			}
		    else if (strcasecmp(curr_filter_operators,"IS")==0) // case-insensitive LIKE
			{
			  // only NULL and NOT NULL are accepted as curr_filter_values
		      __APPEND_TO_QUERY(" %s %s IS %s", start_string, as_field_name, curr_filter_values);
			}
		    else
		    {
		      __APPEND_TO_QUERY(" %s %s %s \'%s\' ", start_string, as_field_name, curr_filter_operators, cou_buildEncodedSQL(cnx, curr_filter_values) );
		    }
	 }
      }

      __SEND_MESG_DEBUG("We are after WHERE")

		if(group_by_class_names != NULL && group_by_prop_names != NULL && get_prop_aggregates!=NULL && coa_GetElementCount(cnx, get_prop_aggregates)>0)
		{ 
			NElems = coa_GetElementCount(cnx, group_by_class_names);
			__ASSERT(NElems>=0, "error getting ElementCount on group_by")
			__ASSERT(NElems == coa_GetElementCount(cnx, group_by_prop_names), "different number of group_by_class_names and group_by_prop_names")

			for(i=0; i<NElems; i++)
			{
				if(i==0)
				{
					snprintf(start_string,10," GROUP BY ");
				}
				if(i==1)
				{
					snprintf(start_string,10, " , \n");
				}
				__APPEND_TO_QUERY(" %s cov_%s_%s.%s ", start_string, par_area, coa_GetStringElement(cnx, group_by_class_names, i), 
				                                                               coa_GetStringElement(cnx, group_by_prop_names, i))
			}
		}

      __SEND_MESG_DEBUG("We are after GROUP BY")

      if(get_prop_aggregates==NULL || coa_GetElementCount(cnx, get_prop_aggregates)==0)
      { 
      if(order_by_class_names != NULL && order_by_prop_names != NULL) 
      {
         more_desc = 0;
         if(order_by_desc != NULL) more_desc = 1;
         NElems = coa_GetElementCount(cnx, order_by_class_names);
         for(i=0; i<NElems; i++)
         {
            char * curr_order_by_class_names = coa_GetElement(cnx, order_by_class_names, i);
            char * curr_order_by_prop_names  = coa_GetElement(cnx, order_by_prop_names, i);
            char * curr_order_by_desc        = coa_GetElement(cnx, order_by_desc, i);

            if(i==0)
            {
              snprintf(start_string,10," ORDER BY ");
            }
            if(i==1)
            {
              snprintf(start_string,10, " , \n");
            }
	    if (*curr_order_by_class_names) {
        	__APPEND_TO_QUERY("\n %s cov_%s_%s.%s ", start_string, par_area, curr_order_by_class_names, curr_order_by_prop_names);
	    } else {
		__APPEND_TO_QUERY("\n %s \"%s\" ", start_string, curr_order_by_prop_names);
	    }

            if(more_desc && curr_order_by_desc == NULL) more_desc=0;
            if(more_desc && curr_order_by_desc != NULL && strcmp(curr_order_by_desc,"1")==0)
            {
               __APPEND_TO_QUERY(" DESC ");
            }
         }
      }

      if(limit>0)  { __APPEND_TO_QUERY(" LIMIT %d ", limit); }
      if(offset>0) { __APPEND_TO_QUERY(" OFFSET %d ", offset); }
      } // if(get_prop_aggregates!=NULL)

      __SEND_MESG_DEBUG("SELECT SQL: %s ", SQLString)

      __GET_RESULT_AND_CREATE_JS
      // We set field names "manualy" to be sure that they are not cutted to POstgres limit size

      if(strcmp("*", coa_GetElement(cnx, get_prop_names, 0))!=0)
      {
        NGetNames = coa_GetElementCount(cnx, get_class_names);
        for(i=0; i<NGetNames; i++)
        {
           cor_SetFieldName(cnx, corec, i, ap_pstrdup(p, coa_GetElement(cnx, get_prop_names, i)));
        }
      }
      
      %%
      ;
      
    };
  };


  method getRelatedClasses
   {
      doc "retrieves names of related classes, foreign column names and relation names for given class";

     id_permission __PERMISSION_GET_RELATED;
     sec_type __METHOD_SEC_TYPE_CLASS;
     
    __PARAMS_HEADER_WITH_CALLBACK

       body language_c
        {
          C
	  %%
	  __CREATE_QUERY("
SELECT P.class_name, P.prop_name, P.prop_foreign_prop, P.prop_foreign_dynamic_prop, 
(SELECT text FROM co_sys_label_relation WHERE id_meta_prop=P.id AND language='%s') AS relation, 
(SELECT text FROM co_sys_label_prop WHERE id_meta_prop=P.id AND language='%s') AS label,
(SELECT text FROM co_sys_label_class AS L, co_sys_meta_class AS C WHERE C.area='%s' AND C.class_name=P.class_name AND L.id_meta_class=C.id AND L.language='%s') AS class_label 
FROM co_sys_meta_prop AS P WHERE prop_foreign_area=\'%s\' AND (prop_foreign=\'%s\' OR (prop_foreign_dynamic=1 AND prop_foreign_dynamic_classes LIKE \'%%|%s|%%\'))
ORDER BY P.class_name
",
	  cnx->language,
	  cnx->language,
	  par_area, cnx->language, par_area,
	  par_class, par_class)

          __GET_RESULT_AND_CREATE_JS
          %%
          ;
       };
   };


 method getSchema
   {
      doc  "Retrieves schema for given class";

    id_permission __PERMISSION_GET;
    sec_type __METHOD_SEC_TYPE_CLASS;
    
    __PARAMS_HEADER_WITH_CALLBACK

    param representant_level
    {
       doc "We are taking only these props that have representant > representat_level"
       {
          pl "Bierzemy tylko te pola, ktore maja representant > representat_level";
       };
       type CO_INTEGER;
       optional 1;
       default "representant_level=-1;";
    };

      body language_c
       {
          C
	  %%
          __CREATE_QUERY("
SELECT * , prop_group_path AS get_group_path, prop_name AS get_prop_name, 
(SELECT text FROM co_sys_label_prop WHERE id_meta_prop=P.id AND language='%s') AS label,
(SELECT text FROM co_sys_label_class AS L, co_sys_meta_class AS C WHERE C.area='%s' AND C.class_name='%s' AND L.id_meta_class=C.id AND L.language='%s') AS class_label 
FROM co_sys_meta_prop AS P
 WHERE area=\'%s\' AND class_name=\'%s\' AND prop_representant > %d AND deleted=0 AND state = 'INSYNC' ORDER BY prop_position, prop_group_path",
					cnx->language,
					par_area,par_class,cnx->language,
					par_area,par_class, representant_level)

          __GET_RESULT_AND_CREATE_JS
	  %%
	  ;
       };
   };

 local_method loc_getSchemaX
 {
   C
    %%
     int loc_getSchemaX
            (
             co_context * cnx, 
             co_recordset * out_corec,
             const char * par_area, 
             const char * par_class,
	     int par_id,
             int deep_level,
             int cur_level,
             const char * get_prop_base,
             const char * base_group_path
            )
    {
      int n_prop_foreign = 0;
      int n_prop_foreign_prop = 0;
      int n_prop_foreign_dynamic = 0;
      int n_prop_foreign_dynamic_classes = 0;
      int n_prop_foreign_dynamic_prop = 0;
      int n_prop_name = 0;
      int n_get_group_path = 0;
      char * prop_foreign;
      char * prop_foreign_prop;
      char * prop_foreign_dynamic;
      char * prop_foreign_dynamic_classes;
      char * prop_foreign_dynamic_prop;
      char * dynamic_class_name;
      char * prop_name;
      char * get_group_path;
      char *ret_js_out;
      char *get_str;
      int prop_representant = 0;
      int nTuples = 0;
      int nProps = 0;
      PGresult * res;
      request_rec *r = cnx->request;
      pool *p = r->pool;  
      int err = 0;
      int i = 0;
      char SQLString[SQLBUFSIZE];
      PGconn *conn = cnx->dbconn;
      int ok;
      int dirty_res = 0;
      const char *get_prop_base_local = "";
      co_recordset * corec;

      if(deep_level>cur_level)
      { 
         __CREATE_QUERY(
"SELECT *, '%s' || prop_group_path AS get_group_path, '%s>' || prop_name AS get_prop_name ,
(SELECT text FROM co_sys_label_prop WHERE id_meta_prop=P.id AND language='%s') AS label, " 
"(SELECT text FROM co_sys_label_class AS L, co_sys_meta_class AS C WHERE C.area='%s' AND C.class_name='%s' AND L.id_meta_class=C.id AND L.language='%s') AS class_label "
" FROM co_sys_meta_prop AS P WHERE P.area=\'%s\' AND P.class_name=\'%s\' AND P.prop_representant > %d AND P.deleted=0 AND P.state = 'INSYNC' ORDER BY prop_position, prop_group_path", 
					base_group_path,
					get_prop_base,
					cnx->language,
					par_area,par_class,cnx->language,
					par_area,par_class,cur_level)

         __SQL_GET_RESULT

         nTuples             = cor_GetTupleCount(cnx, corec);
         nProps              = cor_GetFieldCount(cnx, corec);
         n_prop_foreign      = cor_GetFieldNum  (cnx, corec, "prop_foreign");
         n_prop_foreign_prop = cor_GetFieldNum  (cnx, corec, "prop_foreign_prop");
         n_prop_foreign_dynamic = cor_GetFieldNum  (cnx, corec, "prop_foreign_dynamic");
         n_prop_foreign_dynamic_classes = cor_GetFieldNum  (cnx, corec, "prop_foreign_dynamic_classes");
         n_prop_foreign_dynamic_prop = cor_GetFieldNum  (cnx, corec, "prop_foreign_dynamic_prop");
         n_prop_name         = cor_GetFieldNum  (cnx, corec, "prop_name");
         n_get_group_path   = cor_GetFieldNum  (cnx, corec, "get_group_path");

         for(i=0; i<nTuples; i++)
         {
            cor_AppendRecord(cnx, out_corec, corec, i);

            prop_foreign      = cor_GetCopyOfValue(cnx, corec, i, n_prop_foreign);
            prop_foreign_prop = cor_GetCopyOfValue(cnx, corec, i, n_prop_foreign_prop);
            prop_foreign_dynamic = cor_GetCopyOfValue(cnx, corec, i, n_prop_foreign_dynamic);
            prop_foreign_dynamic_classes = cor_GetCopyOfValue(cnx, corec, i, n_prop_foreign_dynamic_classes);
            prop_foreign_dynamic_prop = cor_GetCopyOfValue(cnx, corec, i, n_prop_foreign_dynamic_prop);
            prop_name         = cor_GetCopyOfValue(cnx, corec, i, n_prop_name);
            get_group_path   = cor_GetCopyOfValue(cnx, corec, i, n_get_group_path);

            if( strlen(prop_foreign) >0)
            {
               get_prop_base_local = ap_psprintf(p, "%s>%s>%s>%s", get_prop_base, prop_name, prop_foreign_prop, prop_foreign);
               loc_getSchemaX(cnx, out_corec, par_area, prop_foreign, par_id, deep_level, cur_level+1, get_prop_base_local, 
								get_group_path);
             }
            if(!strcmp(prop_foreign_dynamic,"1"))
            {
		@co_getProp(par_area, par_class, par_id, prop_foreign_dynamic_prop);
//		if(err==0) dynamic_class_name = atoi(get_str);

    		get_prop_base = ap_psprintf(p, "%s>%s>%s", prop_name, prop_foreign_prop, get_str);
        	cou_sendMessageX(cnx, ap_psprintf(p, "get_prop_base %s", get_prop_base), CO_MSG_DEBUG);
        	cou_sendMessageX(cnx, ap_psprintf(p, "executing loc_getSchemaX"), CO_MSG_DEBUG);
		loc_getSchemaX(cnx, out_corec, par_area, get_str, par_id, deep_level, cur_level+1,get_prop_base_local, get_group_path);
        	cou_sendMessageX(cnx, ap_psprintf(p, "finished execution loc_getSchemaX"), CO_MSG_DEBUG);
	    }
          }
        __CLEAR_RESULT
       } // if
      return 1;
    }
    %%
    ;
 };

 method getSchemaX
   {
      doc  "Retrieves schema for given class";

    id_permission __PERMISSION_GET;
    sec_type __METHOD_SEC_TYPE_CLASS;
    
    __PARAMS_HEADER_WITH_CALLBACK

    param deep_level
    {
       doc "This is max number of levels to go up to container"
       {
          pl "To jest maksymalna liczba odwoa do klas pojemnikw";
       };
       type CO_INTEGER;
       optional 1;
       default "deep_level=3";
    };


     declare language_c
      {
        C "int nTuples = 0;";
        C "int nProps = 0;";
        C "int n_prop_foreign = 0;";
        C "int n_prop_foreign_prop = 0;";
        C "int n_prop_foreign_dynamic = 0;";
        C "int n_prop_foreign_dynamic_classes = 0;";
        C "int n_prop_foreign_dynamic_prop = 0;";
        C "int n_prop_name = 0;";
        C "int n_prop_group_path = 0;";
        C "const char * prop_foreign;";
        C "const char * prop_foreign_prop;";
        C "const char * prop_foreign_dynamic;";
        C "const char * prop_foreign_dynamic_classes;";
        C "const char * prop_foreign_dynamic_prop;";
        C "const char * dynamic_class_name;";
        C "const char * prop_name;";
        C "const char * prop_val;";
        C "const char * prop_group_path;";
        C "int i=0;";
        C 'int in_js_num = 0;';
        C 'const char * get_prop_base = "";';
        C 'int cur_level = 0;';
        C 'co_recordset *out_corec;';
      };

      body language_c
       {
          C
	  %%
          out_corec = cor_CreateEmptyNamed(cnx, "corec2 for getSchemaX");
          __CREATE_QUERY("
SELECT * , prop_group_path AS get_group_path, prop_name AS get_prop_name ,
(SELECT text FROM co_sys_label_prop WHERE id_meta_prop=P.id AND language='%s') AS label,
(SELECT text FROM co_sys_label_class AS L, co_sys_meta_class AS C WHERE C.area='%s' AND C.class_name='%s' AND L.id_meta_class=C.id AND L.language='%s') AS class_label
 FROM co_sys_meta_prop AS P WHERE P.area=\'%s\' AND P.class_name=\'%s\' AND P.prop_representant > %d AND P.deleted=0 AND P.state = 'INSYNC' ORDER BY prop_position , prop_group_path", 
                    cnx->language,
                    par_area, par_class, cnx->language,
					par_area, par_class, -1)

	  __SQL_GET_RESULT
         __ASSERT(ok, "Could not get properties in getSchemaX, par_area: '%s', par_class: '%s'", par_area, par_class)

         nTuples             = cor_GetTupleCount(cnx, corec);
         nProps              = cor_GetFieldCount(cnx, corec);
         n_prop_foreign      = cor_GetFieldNum  (cnx, corec, "prop_foreign");
         n_prop_foreign_prop = cor_GetFieldNum  (cnx, corec, "prop_foreign_prop");
         n_prop_foreign_dynamic = cor_GetFieldNum  (cnx, corec, "prop_foreign_dynamic");
         n_prop_foreign_dynamic_classes = cor_GetFieldNum  (cnx, corec, "prop_foreign_dynamic_classes");
         n_prop_foreign_dynamic_prop = cor_GetFieldNum  (cnx, corec, "prop_foreign_dynamic_prop");
         n_prop_name         = cor_GetFieldNum  (cnx, corec, "prop_name");
         n_prop_group_path        = cor_GetFieldNum  (cnx, corec, "get_group_path");

         for(i=0; i<nTuples; i++)
         {
            cor_AppendRecord(cnx, out_corec, corec, i);
            prop_foreign      = cor_GetCopyOfValue(cnx, corec, i, n_prop_foreign);
            prop_foreign_prop = cor_GetCopyOfValue(cnx, corec, i, n_prop_foreign_prop);
            prop_foreign_dynamic = cor_GetCopyOfValue(cnx, corec, i, n_prop_foreign_dynamic);
            prop_foreign_dynamic_classes = cor_GetCopyOfValue(cnx, corec, i, n_prop_foreign_dynamic_classes);
            prop_foreign_dynamic_prop = cor_GetCopyOfValue(cnx, corec, i, n_prop_foreign_dynamic_prop);
	    prop_name         = cor_GetCopyOfValue(cnx, corec, i, n_prop_name);
	    prop_group_path   = cor_GetCopyOfValue(cnx, corec, i, n_prop_group_path);

            if(strlen(prop_foreign) >0)
            {
                get_prop_base = ap_psprintf(p, "%s>%s>%s", prop_name, prop_foreign_prop, prop_foreign);
                loc_getSchemaX(cnx, out_corec, par_area, prop_foreign, par_id, deep_level, cur_level,get_prop_base, prop_group_path);
            }

            if(!strcmp(prop_foreign_dynamic,"1"))
            {
		@co_getProp(par_area, par_class, par_id, prop_foreign_dynamic_prop);
//		if(err==0) dynamic_class_name = atoi(get_str);


    		get_prop_base = ap_psprintf(p, "%s>%s>%s", prop_name, prop_foreign_prop, get_str);
        	cou_sendMessageX(cnx, ap_psprintf(p, "get_prop_base %s", get_prop_base), CO_MSG_DEBUG);
        	cou_sendMessageX(cnx, ap_psprintf(p, "executing loc_getSchemaX"), CO_MSG_DEBUG);
		loc_getSchemaX(cnx, out_corec, par_area, get_str, par_id, deep_level, cur_level,get_prop_base, prop_group_path);
        	cou_sendMessageX(cnx, ap_psprintf(p, "finished execution loc_getSchemaX"), CO_MSG_DEBUG);
	    }

            cou_sendMessageX(cnx, ap_psprintf(p, "finished %d", i), CO_MSG_DEBUG);
          }

          __CREATE_JS_FROM(out_corec)
          cou_setReturnValue(cnx, out_corec, CO_RECORDSET);
          __CLEAR_RESULT          
	  %%
	  ;
       };
   };

	method showFormFind
	{

		id_permission __PERMISSION_SELECT;
		sec_type __METHOD_SEC_TYPE_CLASS;

		__PARAMS_HEADER
     
		param id_report
		{
			doc "If this param is set we read seetings for the report, if not we create default report for current class";
			type CO_INTEGER;
			optional 1;
			default "id_report=0;";
		};

		param report_name
		{
			doc "If this param is set we read seetings for the report looking for it by name, if not we create default report for current class";
			doc "When name and id is given we use id";
			type CO_STRING;
			optional 1;
			default "report_name = ap_pstrdup(p, \"\");";
		};


		param context_name
		{
			doc "Name of context in which we execute this form. Now this is the name of additionall menu";
			type CO_STRING;
			optional 1;
			default "context_name=ap_pstrdup(p, \"\");";
		};

		param context1
		{
			doc "Additionall parameter passed to menu";
			type CO_STRING;
			optional 1;
			default "context1=ap_pstrdup(p, \"\");";
		};

		param context2
		{
			doc "Additionall parameter passed to menu";
			type CO_STRING;
			optional 1;
			default "context2=ap_pstrdup(p, \"\");";
		};

		param context3
		{
			doc "Additionall parameter passed to menu";
			type CO_STRING;
			optional 1;
			default "context3=ap_pstrdup(p, \"\");";
		};

		param default_open_class_name
		{
			doc "Name of the class to be used to open element";
			type CO_STRING;
			optional 1;
			default "default_open_class_name=par_class;";
		};

		param default_open_id_prop_name
		{
			doc "Name of the prop with id value to be used to open element";
			type CO_STRING;
			optional 1;
			default "default_open_id_prop_name=ap_pstrdup(p, \"id\");";
		};

		param filter_variables
		{
			doc "Array of filter variables in form [name, value]";
			type CO_ARRAY_OF_STRINGS;
			optional 1;
			default "filter_variables=NULL;";
		};




		declare language_c
		{
			C
			%%
				FILE *fh;
				int c;
				const char* root_path;
				long len;
				int user_id=1;
				char* name_role = NULL;
			%%
			;
      };

		body language_c
		{
			C
			%%
			root_path = cou_getCORootPath( cnx );

			__SQL_GET_STRING(name_role, "SELECT name FROM co_sys_role WHERE id=%d", cnx->id_role)

			__SEND_MESG_DEBUG("we have name_role: %s", name_role)

			if(id_report == 0 && strlen(report_name)>0)
			{
				@co_getPropByName('"sys"', '"report"', report_name, '"id"');
				if(err==0)
				{
					id_report = atoi(get_str);
				}
			}

			fh = ap_pfopen(p, ap_pstrcat( p, root_path, "/forms/", "generic_f", "_head.html", NULL), "r");
			if (!fh) cou_error( cnx, "Head file not found");
			cou_logError( cnx, CO_LOG_DEBUG, "sending head" );
			ap_rputs("<HEAD>\n", r);
			ap_kill_timeout(r);

			len = ap_send_fd( fh, r ); /* returns total of bytes sent */
			ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "Head sent: %ld bytes", len);

			ap_pfclose(p,fh);

			cou_logError( cnx, CO_LOG_DEBUG, "sending script" );
			ap_rputs("<SCRIPT>\n", r);

			ap_rprintf( r, "var  id_session = %d;\n", cou_assignNewSession( cnx ) );
			ap_rprintf( r, "var  par_id = %d;\n", par_id );
			ap_rprintf( r, "var  par_class=\'%s\';\n", par_class );
			ap_rprintf( r, "var  par_area = \'%s\';\n", par_area );
			ap_rprintf( r, "var  debug_level = %d;\n", cnx->debug_level );
			ap_rprintf( r, "var  id_report = %d;\n", id_report );
			ap_rprintf( r, "var  context_name = \'%s\';\n", context_name );
			ap_rprintf( r, "var  context1 = \'%s\';\n", context1 );
			ap_rprintf( r, "var  context2 = \'%s\';\n", context2 );
			ap_rprintf( r, "var  context3 = \'%s\';\n", context3 );

			ap_rprintf( r, "var  default_open_class_name = \'%s\';\n", default_open_class_name);
			ap_rprintf( r, "var  default_open_id_prop_name = \'%s\';\n", default_open_id_prop_name);

			ap_rprintf( r, "var  role = {id : %d , name: \'%s\'};\n", cnx->id_role, name_role );

			if(filter_variables !=NULL)
			{
				int NElems = coa_GetElementCount(cnx, filter_variables);
				int count = 0;
				ap_rprintf( r, "var filter_variables = { ");
				for(count=0; count<NElems; count +=2)
				{
					if(count>0){ ap_rprintf(r, ", "); };
					ap_rprintf( r, " %s : '%s' ", 
					                   coa_GetStringElement(cnx, filter_variables, count), 
					                   coa_GetStringElement(cnx, filter_variables, count+1)
					          );
				}
				ap_rprintf( r, " }; \n");
			}
			else
			{
				ap_rprintf( r, "var filter_variables = new Object(); \n");
			}	
			ap_rputs("</SCRIPT>\n", r);

			ap_rputs("</HEAD>\n", r);
			fh = ap_pfopen(p, ap_pstrcat( p, root_path, "/forms/", "generic_f", "_body.html", NULL), "r");
			if (!fh) cou_error( cnx, "Body file not found");
			cou_logError( cnx, CO_LOG_DEBUG, "sending body" );
			ap_rputs("<BODY>\n", r);
			len = ap_send_fd( fh, r ); /* returns total of bytes sent */

			if (r->connection->aborted) ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "connection aborted");
			ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "bytes_send=%ld", r->bytes_sent);

			ap_soft_timeout((char *) "finish request", r);

			ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "Body sent: %ld bytes", len);

			ap_pfclose(p,fh);
			ap_rputs("</BODY>\n", r);

			ap_rflush(r);
			%%
			;
		};
	};

};

//END OF CLASS GENERIC
