// TYPES FOR CO

#include "../../include/types.t"
#include "../../include/defines.t"


// CLASS sevent

class sevent
{
  area "sys";
  area "usr";
  area "sec";

 // PROPERTIES

#include "../../include/generic_props.t"

  inherits generic_acl;
  include_c "co_jscript.h";
  include_c "../js_code/co_js_code.h";

  container_class '';
  container_column_name '';

  prop id_sevent_type
  {
   type CO_INTEGER;
   default 0;
   params '';
   representant 1;
   foreign 'sevent_type';
   foreign_prop 'id';
   position 100;
  };

  prop execution_mode
  {
   type CO_STRING;
   default "''";
   params '';
   representant 1;
   foreign '';
   foreign_prop '';
   position 100;
  }

  prop id_js_code
  {
   type CO_INTEGER;
   default 0;
   params '';
   representant 1;
   foreign 'js_code';
   foreign_prop 'id';
   position 100;
  };

 // METHODS
  include_c "../generic/co_generic.h";

  method fire
  {
    id_permission __PERMISSION_EVALUATE;

    __PARAMS_HEADER
    body language_c
    {
     C
     %%
	/* get context info, create prolog (context.id_spreadsheet etc.), use
	    js_code.evaluate to fire event */

     %%
     ;
    };
  };

    method getObjectEvents
    {
    id_permission __PERMISSION_SELECT;

     __PARAMS_HEADER

      param id_spreadsheet { type CO_INTEGER; };
      param id_column { type CO_INTEGER; };
      param id_row { type CO_INTEGER; };

      body language_c
      {
        C
        %%
        // here we should gather events info
        __CREATE_QUERY(
	"SELECT ET.type AS cevtype, SCR.code AS code, E.id_js_code AS id_js_code, E.id AS id, SCR.version AS version "
        "FROM co_%s_%s E, co_%s_cevent_type ET, co_%s_js_code SCR "
        "WHERE (E.event_type_id=ET.id) AND "
	"(E.spreadsheet_id=%d) AND (E.column_id=%d) AND (E.row_id=%d) AND (SCR.id=E.id_js_code)",
	par_area, par_class, par_area, par_area, 
	id_spreadsheet, id_column, id_row)

        __SQL_GET_RESULT

        __RETURN_RESULT

        %%
        ;
      };
    };

  method setEventScript
  {
    id_permission __PERMISSION_UPDATE;

    __PARAMS_HEADER
    param event_script { type CO_STRING; };
    param version{ type CO_INTEGER; };

    declare language_c
    {
    C
    %%
	int id_js_code;
    %%
    ;
    };

    body language_c
    {
    C
    %%
//
        @co_getProp(par_area, par_class, par_id,'"id_js_code"');

	if(err==0)
	{
	    id_js_code = atoi(get_str);
	}
	else
	{
	    cou_sendMessage(cnx, ap_psprintf(p, "could not find event info for id: %d",par_id));
	    return -1;
	};

	@co_call(par_area, '"js_code"', "update", par_area=>par_area, par_class=>'"js_code"', par_id=>id_js_code, par_prop=>'"code"', new_value=>event_script ,version=>version);
    %%
    ;
    };
  };

  method createEventScript
  {
    id_permission __PERMISSION_CREATE;

    __PARAMS_HEADER
    param id_spreadsheet { type CO_INTEGER; };
    param id_column { type CO_INTEGER; };
    param id_row { type CO_INTEGER; };
    param event_type_id { type CO_INTEGER; };
    param event_script { type CO_STRING; };

    declare language_c
    {
    C
    %%
	int id_js_code;
	int id_cevent;
	int *ret_ptr;
	int my_err=0;
    %%
    ;
    };

    body language_c
    {
    C
    %%
//
	@co_call(par_area, '"js_code"', "createScript", par_area=>par_area, par_class=>'"js_code"', par_id=>-1, code=>event_script );
	ret_ptr = cou_getReturnValue( cnx, CO_INTEGER,  &my_err);
	if (!ret_ptr) cou_error( cnx, "Cannot get return value from js_code::createScript");
	id_js_code = *ret_ptr;
	if (!id_js_code) cou_error( cnx, "js_code::createScript returned 0");

    __SQL_GET_INTEGER( id_cevent, "
INSERT INTO co_%s_%s (spreadsheet_id, column_id, row_id, event_type_id, id_js_code)
  VALUES(%d, %d, %d, %d, %d);
SELECT last_value FROM co_%s_%s_sqid;
",      par_area, par_class, 
	id_spreadsheet, id_column, id_row, event_type_id, id_js_code,
        par_area, par_class )

	if (!(id_cevent>0)) cou_error( cnx, "Error while inserting spreadsheet cevent");

    cou_setReturnValue( cnx, &id_cevent, CO_INTEGER );
    cou_logError( cnx, CO_LOG_DEBUG, "Result: %d", id_cevent );

      //setting ACL for current role; should copy from spreadsheet
      __CREATE_QUERY("INSERT INTO co_sec_acl (id_permission, id_role, area, class_name, id_object)
                      VALUES ( %d , %d, '%s', '%s', %d)",
                      __PERMISSION_CHANGE_ACL, cnx->id_role, par_area, "js_code", id_js_code);
      __SQL_EXEC

      __CREATE_QUERY("INSERT INTO co_sec_acl (id_permission, id_role, area, class_name, id_object)
                      VALUES ( %d , %d, '%s', '%s', %d)",
                      __PERMISSION_EVALUATE, cnx->id_role, par_area, "js_code", id_js_code);
      __SQL_EXEC

      __CREATE_QUERY("INSERT INTO co_sec_acl (id_permission, id_role, area, class_name, id_object)
                      VALUES ( %d , %d, '%s', '%s', %d)",
                      __PERMISSION_GET, cnx->id_role, par_area, "js_code", id_js_code);
      __SQL_EXEC

      __CREATE_QUERY("INSERT INTO co_sec_acl (id_permission, id_role, area, class_name, id_object)
                      VALUES ( %d , %d, '%s', '%s', %d)",
                      __PERMISSION_UPDATE, cnx->id_role, par_area, "js_code", id_js_code);
      __SQL_EXEC

      // spreadsheet_cevent
      __CREATE_QUERY("INSERT INTO co_sec_acl (id_permission, id_role, area, class_name, id_object)
                      VALUES ( %d , %d, '%s', '%s', %d)",
                      __PERMISSION_SELECT, cnx->id_role, par_area, "spreadsheet_cevent", id_cevent);
      __SQL_EXEC

      __CREATE_QUERY("INSERT INTO co_sec_acl (id_permission, id_role, area, class_name, id_object)
                      VALUES ( %d , %d, '%s', '%s', %d)",
                      __PERMISSION_UPDATE, cnx->id_role, par_area, "spreadsheet_cevent", id_cevent);
      __SQL_EXEC

      __CREATE_QUERY("INSERT INTO co_sec_acl (id_permission, id_role, area, class_name, id_object)
                      VALUES ( %d , %d, '%s', '%s', %d)",
                      __PERMISSION_GET, cnx->id_role, par_area, "spreadsheet_cevent", id_cevent);
      __SQL_EXEC

          {
            const char* class_names[] = {"spreadsheet","spreadsheet_column","spreadsheet_row",NULL};
            int object_ids[] = {id_spreadsheet,id_column,id_row};

            cou_queueActionX( cnx, "spreadsheet_cevent", class_names, object_ids,
                CO_OP_INSERT, 
		"event_type_id", ap_psprintf(p, "%d", event_type_id), 
		"id_js_code", ap_psprintf(p, "%d", id_js_code),
		"id", ap_psprintf(p, "%d", id_cevent),
		NULL );
          }


    %%
    ;
    };
  };


};

//END OF CLASS spreadsheet_cevent